# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import unittest
import logging
import struct
import inspect
from os_ken.ofproto import ether, inet
from os_ken.lib.packet import arp
from os_ken.lib.packet import bpdu
from os_ken.lib.packet import ethernet
from os_ken.lib.packet import icmp, icmpv6
from os_ken.lib.packet import ipv4, ipv6
from os_ken.lib.packet import llc
from os_ken.lib.packet import packet, packet_utils
from os_ken.lib.packet import sctp
from os_ken.lib.packet import tcp, udp
from os_ken.lib.packet import vlan
from os_ken.lib import addrconv


LOG = logging.getLogger('test_packet')


class TestPacket(unittest.TestCase):
    """ Test case for packet
    """

    dst_mac = 'aa:aa:aa:aa:aa:aa'
    src_mac = 'bb:bb:bb:bb:bb:bb'
    dst_mac_bin = addrconv.mac.text_to_bin(dst_mac)
    src_mac_bin = addrconv.mac.text_to_bin(src_mac)
    dst_ip = '192.168.128.10'
    src_ip = '192.168.122.20'
    dst_ip_bin = addrconv.ipv4.text_to_bin(dst_ip)
    src_port = 50001
    dst_port = 50002
    src_ip_bin = addrconv.ipv4.text_to_bin(src_ip)
    payload = b'\x06\x06\x47\x50\x00\x00\x00\x00' \
        + b'\xcd\xc5\x00\x00\x00\x00\x00\x00' \
        + b'\x10\x11\x12\x13\x14\x15\x16\x17' \
        + b'\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'

    def get_protocols(self, pkt):
        protocols = {}
        for p in pkt:
            if hasattr(p, 'protocol_name'):
                protocols[p.protocol_name] = p
            else:
                protocols['payload'] = p
        return protocols

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_arp(self):
        # buid packet
        e = ethernet.ethernet(self.dst_mac, self.src_mac,
                              ether.ETH_TYPE_ARP)
        a = arp.arp(1, ether.ETH_TYPE_IP, 6, 4, 2,
                    self.src_mac, self.src_ip, self.dst_mac,
                    self.dst_ip)
        p = packet.Packet()
        p.add_protocol(e)
        p.add_protocol(a)
        p.serialize()

        # ethernet !6s6sH
        e_buf = self.dst_mac_bin \
            + self.src_mac_bin \
            + b'\x08\x06'

        # arp !HHBBH6sI6sI
        a_buf = b'\x00\x01' \
            + b'\x08\x00' \
            + b'\x06' \
            + b'\x04' \
            + b'\x00\x02' \
            + self.src_mac_bin \
            + self.src_ip_bin \
            + self.dst_mac_bin \
            + self.dst_ip_bin

        buf = e_buf + a_buf

        # Append padding if ethernet frame is less than 60 bytes length
        pad_len = 60 - len(buf)
        if pad_len > 0:
            buf += b'\x00' * pad_len
        self.assertEqual(buf, p.data)

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_arp = protocols['arp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual(self.dst_mac, p_eth.dst)
        self.assertEqual(self.src_mac, p_eth.src)
        self.assertEqual(ether.ETH_TYPE_ARP, p_eth.ethertype)

        # arp
        self.assertTrue(p_arp)
        self.assertEqual(1, p_arp.hwtype)
        self.assertEqual(ether.ETH_TYPE_IP, p_arp.proto)
        self.assertEqual(6, p_arp.hlen)
        self.assertEqual(4, p_arp.plen)
        self.assertEqual(2, p_arp.opcode)
        self.assertEqual(self.src_mac, p_arp.src_mac)
        self.assertEqual(self.src_ip, p_arp.src_ip)
        self.assertEqual(self.dst_mac, p_arp.dst_mac)
        self.assertEqual(self.dst_ip, p_arp.dst_ip)

        # to string
        eth_values = {'dst': self.dst_mac,
                      'src': self.src_mac,
                      'ethertype': ether.ETH_TYPE_ARP}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        arp_values = {'hwtype': 1,
                      'proto': ether.ETH_TYPE_IP,
                      'hlen': 6,
                      'plen': 4,
                      'opcode': 2,
                      'src_mac': self.src_mac,
                      'dst_mac': self.dst_mac,
                      'src_ip': self.src_ip,
                      'dst_ip': self.dst_ip}
        _arp_str = ','.join(['%s=%s' % (k, repr(arp_values[k]))
                             for k, v in inspect.getmembers(p_arp)
                             if k in arp_values])
        arp_str = '%s(%s)' % (arp.arp.__name__, _arp_str)

        pkt_str = '%s, %s' % (eth_str, arp_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(arp_str, str(p_arp))
        self.assertEqual(arp_str, repr(p_arp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_vlan_arp(self):
        # buid packet
        e = ethernet.ethernet(self.dst_mac, self.src_mac,
                              ether.ETH_TYPE_8021Q)
        v = vlan.vlan(0b111, 0b1, 3, ether.ETH_TYPE_ARP)
        a = arp.arp(1, ether.ETH_TYPE_IP, 6, 4, 2,
                    self.src_mac, self.src_ip, self.dst_mac,
                    self.dst_ip)
        p = packet.Packet()
        p.add_protocol(e)
        p.add_protocol(v)
        p.add_protocol(a)
        p.serialize()

        # ethernet !6s6sH
        e_buf = self.dst_mac_bin \
            + self.src_mac_bin \
            + b'\x81\x00'

        # vlan !HH
        v_buf = b'\xF0\x03' \
            + b'\x08\x06'

        # arp !HHBBH6sI6sI
        a_buf = b'\x00\x01' \
            + b'\x08\x00' \
            + b'\x06' \
            + b'\x04' \
            + b'\x00\x02' \
            + self.src_mac_bin \
            + self.src_ip_bin \
            + self.dst_mac_bin \
            + self.dst_ip_bin

        buf = e_buf + v_buf + a_buf

        # Append padding if ethernet frame is less than 60 bytes length
        pad_len = 60 - len(buf)
        if pad_len > 0:
            buf += b'\x00' * pad_len
        self.assertEqual(buf, p.data)

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_vlan = protocols['vlan']
        p_arp = protocols['arp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual(self.dst_mac, p_eth.dst)
        self.assertEqual(self.src_mac, p_eth.src)
        self.assertEqual(ether.ETH_TYPE_8021Q, p_eth.ethertype)

        # vlan
        self.assertTrue(p_vlan)
        self.assertEqual(0b111, p_vlan.pcp)
        self.assertEqual(0b1, p_vlan.cfi)
        self.assertEqual(3, p_vlan.vid)
        self.assertEqual(ether.ETH_TYPE_ARP, p_vlan.ethertype)

        # arp
        self.assertTrue(p_arp)
        self.assertEqual(1, p_arp.hwtype)
        self.assertEqual(ether.ETH_TYPE_IP, p_arp.proto)
        self.assertEqual(6, p_arp.hlen)
        self.assertEqual(4, p_arp.plen)
        self.assertEqual(2, p_arp.opcode)
        self.assertEqual(self.src_mac, p_arp.src_mac)
        self.assertEqual(self.src_ip, p_arp.src_ip)
        self.assertEqual(self.dst_mac, p_arp.dst_mac)
        self.assertEqual(self.dst_ip, p_arp.dst_ip)

        # to string
        eth_values = {'dst': self.dst_mac,
                      'src': self.src_mac,
                      'ethertype': ether.ETH_TYPE_8021Q}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        vlan_values = {'pcp': 0b111,
                       'cfi': 0b1,
                       'vid': 3,
                       'ethertype': ether.ETH_TYPE_ARP}
        _vlan_str = ','.join(['%s=%s' % (k, repr(vlan_values[k]))
                              for k, v in inspect.getmembers(p_vlan)
                              if k in vlan_values])
        vlan_str = '%s(%s)' % (vlan.vlan.__name__, _vlan_str)

        arp_values = {'hwtype': 1,
                      'proto': ether.ETH_TYPE_IP,
                      'hlen': 6,
                      'plen': 4,
                      'opcode': 2,
                      'src_mac': self.src_mac,
                      'dst_mac': self.dst_mac,
                      'src_ip': self.src_ip,
                      'dst_ip': self.dst_ip}
        _arp_str = ','.join(['%s=%s' % (k, repr(arp_values[k]))
                             for k, v in inspect.getmembers(p_arp)
                             if k in arp_values])
        arp_str = '%s(%s)' % (arp.arp.__name__, _arp_str)

        pkt_str = '%s, %s, %s' % (eth_str, vlan_str, arp_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(vlan_str, str(p_vlan))
        self.assertEqual(vlan_str, repr(p_vlan))

        self.assertEqual(arp_str, str(p_arp))
        self.assertEqual(arp_str, repr(p_arp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv4_udp(self):
        # buid packet
        e = ethernet.ethernet(self.dst_mac, self.src_mac,
                              ether.ETH_TYPE_IP)
        ip = ipv4.ipv4(4, 5, 1, 0, 3, 1, 4, 64, inet.IPPROTO_UDP, 0,
                       self.src_ip, self.dst_ip)
        u = udp.udp(0x190F, 0x1F90, 0, 0)

        p = packet.Packet()
        p.add_protocol(e)
        p.add_protocol(ip)
        p.add_protocol(u)
        p.add_protocol(self.payload)
        p.serialize()

        # ethernet !6s6sH
        e_buf = self.dst_mac_bin \
            + self.src_mac_bin \
            + b'\x08\x00'

        # ipv4 !BBHHHBBHII
        ip_buf = b'\x45' \
            + b'\x01' \
            + b'\x00\x3C' \
            + b'\x00\x03' \
            + b'\x20\x04' \
            + b'\x40' \
            + b'\x11' \
            + b'\x00\x00' \
            + self.src_ip_bin \
            + self.dst_ip_bin

        # udp !HHHH
        u_buf = b'\x19\x0F' \
            + b'\x1F\x90' \
            + b'\x00\x28' \
            + b'\x00\x00'

        buf = e_buf + ip_buf + u_buf + self.payload

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv4 = protocols['ipv4']
        p_udp = protocols['udp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual(self.dst_mac, p_eth.dst)
        self.assertEqual(self.src_mac, p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IP, p_eth.ethertype)

        # ipv4
        self.assertTrue(p_ipv4)
        self.assertEqual(4, p_ipv4.version)
        self.assertEqual(5, p_ipv4.header_length)
        self.assertEqual(1, p_ipv4.tos)
        l = len(ip_buf) + len(u_buf) + len(self.payload)
        self.assertEqual(l, p_ipv4.total_length)
        self.assertEqual(3, p_ipv4.identification)
        self.assertEqual(1, p_ipv4.flags)
        self.assertEqual(64, p_ipv4.ttl)
        self.assertEqual(inet.IPPROTO_UDP, p_ipv4.proto)
        self.assertEqual(self.src_ip, p_ipv4.src)
        self.assertEqual(self.dst_ip, p_ipv4.dst)
        t = bytearray(ip_buf)
        struct.pack_into('!H', t, 10, p_ipv4.csum)
        self.assertEqual(packet_utils.checksum(t), 0)

        # udp
        self.assertTrue(p_udp)
        self.assertEqual(0x190f, p_udp.src_port)
        self.assertEqual(0x1F90, p_udp.dst_port)
        self.assertEqual(len(u_buf) + len(self.payload), p_udp.total_length)
        self.assertEqual(0x77b2, p_udp.csum)
        t = bytearray(u_buf)
        struct.pack_into('!H', t, 6, p_udp.csum)
        ph = struct.pack('!4s4sBBH', self.src_ip_bin, self.dst_ip_bin, 0,
                         17, len(u_buf) + len(self.payload))
        t = ph + t + self.payload
        self.assertEqual(packet_utils.checksum(t), 0)

        # payload
        self.assertTrue('payload' in protocols)
        self.assertEqual(self.payload, protocols['payload'])

        # to string
        eth_values = {'dst': self.dst_mac,
                      'src': self.src_mac,
                      'ethertype': ether.ETH_TYPE_IP}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv4_values = {'version': 4,
                       'header_length': 5,
                       'tos': 1,
                       'total_length': l,
                       'identification': 3,
                       'flags': 1,
                       'offset': p_ipv4.offset,
                       'ttl': 64,
                       'proto': inet.IPPROTO_UDP,
                       'csum': p_ipv4.csum,
                       'src': self.src_ip,
                       'dst': self.dst_ip,
                       'option': None}
        _ipv4_str = ','.join(['%s=%s' % (k, repr(ipv4_values[k]))
                              for k, v in inspect.getmembers(p_ipv4)
                              if k in ipv4_values])
        ipv4_str = '%s(%s)' % (ipv4.ipv4.__name__, _ipv4_str)

        udp_values = {'src_port': 0x190f,
                      'dst_port': 0x1F90,
                      'total_length': len(u_buf) + len(self.payload),
                      'csum': 0x77b2}
        _udp_str = ','.join(['%s=%s' % (k, repr(udp_values[k]))
                             for k, v in inspect.getmembers(p_udp)
                             if k in udp_values])
        udp_str = '%s(%s)' % (udp.udp.__name__, _udp_str)

        pkt_str = '%s, %s, %s, %s' % (eth_str, ipv4_str, udp_str,
                                      repr(protocols['payload']))

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv4_str, str(p_ipv4))
        self.assertEqual(ipv4_str, repr(p_ipv4))

        self.assertEqual(udp_str, str(p_udp))
        self.assertEqual(udp_str, repr(p_udp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv4_tcp(self):
        # buid packet
        e = ethernet.ethernet(self.dst_mac, self.src_mac,
                              ether.ETH_TYPE_IP)
        ip = ipv4.ipv4(4, 5, 0, 0, 0, 0, 0, 64, inet.IPPROTO_TCP, 0,
                       self.src_ip, self.dst_ip)
        t = tcp.tcp(0x190F, 0x1F90, 0x123, 1, 6, 0b101010, 2048, 0, 0x6f,
                    b'\x01\x02')

        p = packet.Packet()
        p.add_protocol(e)
        p.add_protocol(ip)
        p.add_protocol(t)
        p.add_protocol(self.payload)
        p.serialize()

        # ethernet !6s6sH
        e_buf = self.dst_mac_bin \
            + self.src_mac_bin \
            + b'\x08\x00'

        # ipv4 !BBHHHBBHII
        ip_buf = b'\x45' \
            + b'\x00' \
            + b'\x00\x4C' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x40' \
            + b'\x06' \
            + b'\x00\x00' \
            + self.src_ip_bin \
            + self.dst_ip_bin

        # tcp !HHIIBBHHH + option
        t_buf = b'\x19\x0F' \
            + b'\x1F\x90' \
            + b'\x00\x00\x01\x23' \
            + b'\x00\x00\x00\x01' \
            + b'\x60' \
            + b'\x2A' \
            + b'\x08\x00' \
            + b'\x00\x00' \
            + b'\x00\x6F' \
            + b'\x01\x02\x00\x00'

        buf = e_buf + ip_buf + t_buf + self.payload

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv4 = protocols['ipv4']
        p_tcp = protocols['tcp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual(self.dst_mac, p_eth.dst)
        self.assertEqual(self.src_mac, p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IP, p_eth.ethertype)

        # ipv4
        self.assertTrue(p_ipv4)
        self.assertEqual(4, p_ipv4.version)
        self.assertEqual(5, p_ipv4.header_length)
        self.assertEqual(0, p_ipv4.tos)
        l = len(ip_buf) + len(t_buf) + len(self.payload)
        self.assertEqual(l, p_ipv4.total_length)
        self.assertEqual(0, p_ipv4.identification)
        self.assertEqual(0, p_ipv4.flags)
        self.assertEqual(64, p_ipv4.ttl)
        self.assertEqual(inet.IPPROTO_TCP, p_ipv4.proto)
        self.assertEqual(self.src_ip, p_ipv4.src)
        self.assertEqual(self.dst_ip, p_ipv4.dst)
        t = bytearray(ip_buf)
        struct.pack_into('!H', t, 10, p_ipv4.csum)
        self.assertEqual(packet_utils.checksum(t), 0)

        # tcp
        self.assertTrue(p_tcp)
        self.assertEqual(0x190f, p_tcp.src_port)
        self.assertEqual(0x1F90, p_tcp.dst_port)
        self.assertEqual(0x123, p_tcp.seq)
        self.assertEqual(1, p_tcp.ack)
        self.assertEqual(6, p_tcp.offset)
        self.assertEqual(0b101010, p_tcp.bits)
        self.assertEqual(2048, p_tcp.window_size)
        self.assertEqual(0x6f, p_tcp.urgent)
        self.assertEqual(len(t_buf), len(p_tcp))
        t = bytearray(t_buf)
        struct.pack_into('!H', t, 16, p_tcp.csum)
        ph = struct.pack('!4s4sBBH', self.src_ip_bin, self.dst_ip_bin, 0,
                         6, len(t_buf) + len(self.payload))
        t = ph + t + self.payload
        self.assertEqual(packet_utils.checksum(t), 0)

        # payload
        self.assertTrue('payload' in protocols)
        self.assertEqual(self.payload, protocols['payload'])

        # to string
        eth_values = {'dst': self.dst_mac,
                      'src': self.src_mac,
                      'ethertype': ether.ETH_TYPE_IP}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv4_values = {'version': 4,
                       'header_length': 5,
                       'tos': 0,
                       'total_length': l,
                       'identification': 0,
                       'flags': 0,
                       'offset': p_ipv4.offset,
                       'ttl': 64,
                       'proto': inet.IPPROTO_TCP,
                       'csum': p_ipv4.csum,
                       'src': self.src_ip,
                       'dst': self.dst_ip,
                       'option': None}
        _ipv4_str = ','.join(['%s=%s' % (k, repr(ipv4_values[k]))
                              for k, v in inspect.getmembers(p_ipv4)
                              if k in ipv4_values])
        ipv4_str = '%s(%s)' % (ipv4.ipv4.__name__, _ipv4_str)

        tcp_values = {'src_port': 0x190f,
                      'dst_port': 0x1F90,
                      'seq': 0x123,
                      'ack': 1,
                      'offset': 6,
                      'bits': 0b101010,
                      'window_size': 2048,
                      'csum': p_tcp.csum,
                      'urgent': 0x6f,
                      'option': p_tcp.option}
        _tcp_str = ','.join(['%s=%s' % (k, repr(tcp_values[k]))
                             for k, v in inspect.getmembers(p_tcp)
                             if k in tcp_values])
        tcp_str = '%s(%s)' % (tcp.tcp.__name__, _tcp_str)

        pkt_str = '%s, %s, %s, %s' % (eth_str, ipv4_str, tcp_str,
                                      repr(protocols['payload']))

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv4_str, str(p_ipv4))
        self.assertEqual(ipv4_str, repr(p_ipv4))

        self.assertEqual(tcp_str, str(p_tcp))
        self.assertEqual(tcp_str, repr(p_tcp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv4_sctp(self):
        # build packet
        e = ethernet.ethernet()
        ip = ipv4.ipv4(proto=inet.IPPROTO_SCTP)
        s = sctp.sctp(chunks=[sctp.chunk_data(payload_data=self.payload)])

        p = e / ip / s
        p.serialize()

        ipaddr = addrconv.ipv4.text_to_bin('0.0.0.0')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x08\x00'

        # ipv4 !BBHHHBBHII
        ip_buf = b'\x45' \
            + b'\x00' \
            + b'\x00\x50' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\xff' \
            + b'\x84' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # sctp !HHII + chunk_data !BBHIHHI + payload
        s_buf = b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00' \
            + b'\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + self.payload

        buf = e_buf + ip_buf + s_buf

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv4 = protocols['ipv4']
        p_sctp = protocols['sctp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IP, p_eth.ethertype)

        # ipv4
        self.assertTrue(p_ipv4)
        self.assertEqual(4, p_ipv4.version)
        self.assertEqual(5, p_ipv4.header_length)
        self.assertEqual(0, p_ipv4.tos)
        l = len(ip_buf) + len(s_buf)
        self.assertEqual(l, p_ipv4.total_length)
        self.assertEqual(0, p_ipv4.identification)
        self.assertEqual(0, p_ipv4.flags)
        self.assertEqual(255, p_ipv4.ttl)
        self.assertEqual(inet.IPPROTO_SCTP, p_ipv4.proto)
        self.assertEqual('10.0.0.1', p_ipv4.src)
        self.assertEqual('10.0.0.2', p_ipv4.dst)
        t = bytearray(ip_buf)
        struct.pack_into('!H', t, 10, p_ipv4.csum)
        self.assertEqual(packet_utils.checksum(t), 0x1403)

        # sctp
        self.assertTrue(p_sctp)
        self.assertEqual(1, p_sctp.src_port)
        self.assertEqual(1, p_sctp.dst_port)
        self.assertEqual(0, p_sctp.vtag)
        assert isinstance(p_sctp.chunks[0], sctp.chunk_data)
        self.assertEqual(0, p_sctp.chunks[0]._type)
        self.assertEqual(0, p_sctp.chunks[0].unordered)
        self.assertEqual(0, p_sctp.chunks[0].begin)
        self.assertEqual(0, p_sctp.chunks[0].end)
        self.assertEqual(16 + len(self.payload), p_sctp.chunks[0].length)
        self.assertEqual(0, p_sctp.chunks[0].tsn)
        self.assertEqual(0, p_sctp.chunks[0].sid)
        self.assertEqual(0, p_sctp.chunks[0].seq)
        self.assertEqual(0, p_sctp.chunks[0].payload_id)
        self.assertEqual(self.payload, p_sctp.chunks[0].payload_data)
        self.assertEqual(len(s_buf), len(p_sctp))

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IP}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv4_values = {'version': 4,
                       'header_length': 5,
                       'tos': 0,
                       'total_length': l,
                       'identification': 0,
                       'flags': 0,
                       'offset': 0,
                       'ttl': 255,
                       'proto': inet.IPPROTO_SCTP,
                       'csum': p_ipv4.csum,
                       'src': '10.0.0.1',
                       'dst': '10.0.0.2',
                       'option': None}
        _ipv4_str = ','.join(['%s=%s' % (k, repr(ipv4_values[k]))
                              for k, v in inspect.getmembers(p_ipv4)
                              if k in ipv4_values])
        ipv4_str = '%s(%s)' % (ipv4.ipv4.__name__, _ipv4_str)

        data_values = {'unordered': 0,
                       'begin': 0,
                       'end': 0,
                       'length': 16 + len(self.payload),
                       'tsn': 0,
                       'sid': 0,
                       'seq': 0,
                       'payload_id': 0,
                       'payload_data': self.payload}
        _data_str = ','.join(['%s=%s' % (k, repr(data_values[k]))
                              for k in sorted(data_values.keys())])
        data_str = '[%s(%s)]' % (sctp.chunk_data.__name__, _data_str)

        sctp_values = {'src_port': 1,
                       'dst_port': 1,
                       'vtag': 0,
                       'csum': repr(p_sctp.csum),
                       'chunks': data_str}
        _sctp_str = ','.join(['%s=%s' % (k, sctp_values[k])
                              for k, _ in inspect.getmembers(p_sctp)
                              if k in sctp_values])
        sctp_str = '%s(%s)' % (sctp.sctp.__name__, _sctp_str)

        pkt_str = '%s, %s, %s' % (eth_str, ipv4_str, sctp_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv4_str, str(p_ipv4))
        self.assertEqual(ipv4_str, repr(p_ipv4))

        self.assertEqual(sctp_str, str(p_sctp))
        self.assertEqual(sctp_str, repr(p_sctp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv4_icmp(self):
        # buid packet
        e = ethernet.ethernet()
        ip = ipv4.ipv4(proto=inet.IPPROTO_ICMP)
        ic = icmp.icmp()

        p = e / ip / ic
        p.serialize()

        ipaddr = addrconv.ipv4.text_to_bin('0.0.0.0')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x08\x00'

        # ipv4 !BBHHHBBHII
        ip_buf = b'\x45' \
            + b'\x00' \
            + b'\x00\x1c' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\xff' \
            + b'\x01' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # icmp !BBH + echo !HH
        ic_buf = b'\x08' \
            + b'\x00' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00'

        buf = e_buf + ip_buf + ic_buf

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv4 = protocols['ipv4']
        p_icmp = protocols['icmp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IP, p_eth.ethertype)

        # ipv4
        self.assertTrue(p_ipv4)
        self.assertEqual(4, p_ipv4.version)
        self.assertEqual(5, p_ipv4.header_length)
        self.assertEqual(0, p_ipv4.tos)
        l = len(ip_buf) + len(ic_buf)
        self.assertEqual(l, p_ipv4.total_length)
        self.assertEqual(0, p_ipv4.identification)
        self.assertEqual(0, p_ipv4.flags)
        self.assertEqual(255, p_ipv4.ttl)
        self.assertEqual(inet.IPPROTO_ICMP, p_ipv4.proto)
        self.assertEqual('10.0.0.1', p_ipv4.src)
        self.assertEqual('10.0.0.2', p_ipv4.dst)
        t = bytearray(ip_buf)
        struct.pack_into('!H', t, 10, p_ipv4.csum)
        self.assertEqual(packet_utils.checksum(t), 0x1403)

        # icmp
        self.assertTrue(p_icmp)
        self.assertEqual(8, p_icmp.type)
        self.assertEqual(0, p_icmp.code)
        self.assertEqual(0, p_icmp.data.id)
        self.assertEqual(0, p_icmp.data.seq)
        self.assertEqual(len(ic_buf), len(p_icmp))
        t = bytearray(ic_buf)
        struct.pack_into('!H', t, 2, p_icmp.csum)
        self.assertEqual(packet_utils.checksum(t), 0)

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IP}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, _ in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv4_values = {'version': 4,
                       'header_length': 5,
                       'tos': 0,
                       'total_length': l,
                       'identification': 0,
                       'flags': 0,
                       'offset': p_ipv4.offset,
                       'ttl': 255,
                       'proto': inet.IPPROTO_ICMP,
                       'csum': p_ipv4.csum,
                       'src': '10.0.0.1',
                       'dst': '10.0.0.2',
                       'option': None}
        _ipv4_str = ','.join(['%s=%s' % (k, repr(ipv4_values[k]))
                              for k, _ in inspect.getmembers(p_ipv4)
                              if k in ipv4_values])
        ipv4_str = '%s(%s)' % (ipv4.ipv4.__name__, _ipv4_str)

        echo_values = {'id': 0,
                       'seq': 0,
                       'data': None}
        _echo_str = ','.join(['%s=%s' % (k, repr(echo_values[k]))
                              for k in sorted(echo_values.keys())])
        echo_str = '%s(%s)' % (icmp.echo.__name__, _echo_str)
        icmp_values = {'type': 8,
                       'code': 0,
                       'csum': p_icmp.csum,
                       'data': echo_str}
        _icmp_str = ','.join(['%s=%s' % (k, icmp_values[k])
                              for k, _ in inspect.getmembers(p_icmp)
                              if k in icmp_values])
        icmp_str = '%s(%s)' % (icmp.icmp.__name__, _icmp_str)

        pkt_str = '%s, %s, %s' % (eth_str, ipv4_str, icmp_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv4_str, str(p_ipv4))
        self.assertEqual(ipv4_str, repr(p_ipv4))

        self.assertEqual(icmp_str, str(p_icmp))
        self.assertEqual(icmp_str, repr(p_icmp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv6_udp(self):
        # build packet
        e = ethernet.ethernet(ethertype=ether.ETH_TYPE_IPV6)
        ip = ipv6.ipv6(nxt=inet.IPPROTO_UDP)
        u = udp.udp()

        p = e / ip / u / self.payload
        p.serialize()

        ipaddr = addrconv.ipv6.text_to_bin('::')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x86\xdd'

        # ipv6 !IHBB16s16s'
        ip_buf = b'\x60\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x11' \
            + b'\xff' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # udp !HHHH
        u_buf = b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x28' \
            + b'\x00\x00'

        buf = e_buf + ip_buf + u_buf + self.payload

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv6 = protocols['ipv6']
        p_udp = protocols['udp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IPV6, p_eth.ethertype)

        # ipv6
        self.assertTrue(p_ipv6)
        self.assertEqual(6, p_ipv6.version)
        self.assertEqual(0, p_ipv6.traffic_class)
        self.assertEqual(0, p_ipv6.flow_label)
        self.assertEqual(len(u_buf) + len(self.payload), p_ipv6.payload_length)
        self.assertEqual(inet.IPPROTO_UDP, p_ipv6.nxt)
        self.assertEqual(255, p_ipv6.hop_limit)
        self.assertEqual('10::10', p_ipv6.src)
        self.assertEqual('20::20', p_ipv6.dst)

        # udp
        self.assertTrue(p_udp)
        self.assertEqual(1, p_udp.src_port)
        self.assertEqual(1, p_udp.dst_port)
        self.assertEqual(len(u_buf) + len(self.payload), p_udp.total_length)
        self.assertEqual(0x2B60, p_udp.csum)
        t = bytearray(u_buf)
        struct.pack_into('!H', t, 6, p_udp.csum)
        ph = struct.pack('!16s16sI3xB', ipaddr, ipaddr,
                         len(u_buf) + len(self.payload), 17)
        t = ph + t + self.payload
        self.assertEqual(packet_utils.checksum(t), 0x62)

        # payload
        self.assertTrue('payload' in protocols)
        self.assertEqual(self.payload, protocols['payload'])

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IPV6}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv6_values = {'version': 6,
                       'traffic_class': 0,
                       'flow_label': 0,
                       'payload_length': len(u_buf) + len(self.payload),
                       'nxt': inet.IPPROTO_UDP,
                       'hop_limit': 255,
                       'src': '10::10',
                       'dst': '20::20',
                       'ext_hdrs': []}
        _ipv6_str = ','.join(['%s=%s' % (k, repr(ipv6_values[k]))
                              for k, v in inspect.getmembers(p_ipv6)
                              if k in ipv6_values])
        ipv6_str = '%s(%s)' % (ipv6.ipv6.__name__, _ipv6_str)

        udp_values = {'src_port': 1,
                      'dst_port': 1,
                      'total_length': len(u_buf) + len(self.payload),
                      'csum': 0x2B60}
        _udp_str = ','.join(['%s=%s' % (k, repr(udp_values[k]))
                             for k, v in inspect.getmembers(p_udp)
                             if k in udp_values])
        udp_str = '%s(%s)' % (udp.udp.__name__, _udp_str)

        pkt_str = '%s, %s, %s, %s' % (eth_str, ipv6_str, udp_str,
                                      repr(protocols['payload']))

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv6_str, str(p_ipv6))
        self.assertEqual(ipv6_str, repr(p_ipv6))

        self.assertEqual(udp_str, str(p_udp))
        self.assertEqual(udp_str, repr(p_udp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv6_tcp(self):
        # build packet
        e = ethernet.ethernet(ethertype=ether.ETH_TYPE_IPV6)
        ip = ipv6.ipv6()
        t = tcp.tcp(option=b'\x01\x02')

        p = e / ip / t / self.payload
        p.serialize()

        ipaddr = addrconv.ipv6.text_to_bin('::')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x86\xdd'

        # ipv6 !IHBB16s16s'
        ip_buf = b'\x60\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x06' \
            + b'\xff' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # tcp !HHIIBBHHH + option
        t_buf = b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x60' \
            + b'\x00' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x01\x02\x00\x00'

        buf = e_buf + ip_buf + t_buf + self.payload

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv6 = protocols['ipv6']
        p_tcp = protocols['tcp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IPV6, p_eth.ethertype)

        # ipv6
        self.assertTrue(p_ipv6)
        self.assertEqual(6, p_ipv6.version)
        self.assertEqual(0, p_ipv6.traffic_class)
        self.assertEqual(0, p_ipv6.flow_label)
        self.assertEqual(len(t_buf) + len(self.payload), p_ipv6.payload_length)
        self.assertEqual(inet.IPPROTO_TCP, p_ipv6.nxt)
        self.assertEqual(255, p_ipv6.hop_limit)
        self.assertEqual('10::10', p_ipv6.src)
        self.assertEqual('20::20', p_ipv6.dst)

        # tcp
        self.assertTrue(p_tcp)
        self.assertEqual(1, p_tcp.src_port)
        self.assertEqual(1, p_tcp.dst_port)
        self.assertEqual(0, p_tcp.seq)
        self.assertEqual(0, p_tcp.ack)
        self.assertEqual(6, p_tcp.offset)
        self.assertEqual(0, p_tcp.bits)
        self.assertEqual(0, p_tcp.window_size)
        self.assertEqual(0, p_tcp.urgent)
        self.assertEqual(len(t_buf), len(p_tcp))
        t = bytearray(t_buf)
        struct.pack_into('!H', t, 16, p_tcp.csum)
        ph = struct.pack('!16s16sI3xB', ipaddr, ipaddr,
                         len(t_buf) + len(self.payload), 6)
        t = ph + t + self.payload
        self.assertEqual(packet_utils.checksum(t), 0x62)

        # payload
        self.assertTrue('payload' in protocols)
        self.assertEqual(self.payload, protocols['payload'])

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IPV6}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv6_values = {'version': 6,
                       'traffic_class': 0,
                       'flow_label': 0,
                       'payload_length': len(t_buf) + len(self.payload),
                       'nxt': inet.IPPROTO_TCP,
                       'hop_limit': 255,
                       'src': '10::10',
                       'dst': '20::20',
                       'ext_hdrs': []}
        _ipv6_str = ','.join(['%s=%s' % (k, repr(ipv6_values[k]))
                              for k, v in inspect.getmembers(p_ipv6)
                              if k in ipv6_values])
        ipv6_str = '%s(%s)' % (ipv6.ipv6.__name__, _ipv6_str)

        tcp_values = {'src_port': 1,
                      'dst_port': 1,
                      'seq': 0,
                      'ack': 0,
                      'offset': 6,
                      'bits': 0,
                      'window_size': 0,
                      'csum': p_tcp.csum,
                      'urgent': 0,
                      'option': p_tcp.option}
        _tcp_str = ','.join(['%s=%s' % (k, repr(tcp_values[k]))
                             for k, v in inspect.getmembers(p_tcp)
                             if k in tcp_values])
        tcp_str = '%s(%s)' % (tcp.tcp.__name__, _tcp_str)

        pkt_str = '%s, %s, %s, %s' % (eth_str, ipv6_str, tcp_str,
                                      repr(protocols['payload']))

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv6_str, str(p_ipv6))
        self.assertEqual(ipv6_str, repr(p_ipv6))

        self.assertEqual(tcp_str, str(p_tcp))
        self.assertEqual(tcp_str, repr(p_tcp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv6_sctp(self):
        # build packet
        e = ethernet.ethernet(ethertype=ether.ETH_TYPE_IPV6)
        ip = ipv6.ipv6(nxt=inet.IPPROTO_SCTP)
        s = sctp.sctp(chunks=[sctp.chunk_data(payload_data=self.payload)])

        p = e / ip / s
        p.serialize()

        ipaddr = addrconv.ipv6.text_to_bin('::')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x86\xdd'

        # ipv6 !IHBB16s16s'
        ip_buf = b'\x60\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x84' \
            + b'\xff' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # sctp !HHII + chunk_data !BBHIHHI + payload
        s_buf = b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00' \
            + b'\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00' \
            + b'\x00\x00\x00\x00' \
            + self.payload

        buf = e_buf + ip_buf + s_buf

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv6 = protocols['ipv6']
        p_sctp = protocols['sctp']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IPV6, p_eth.ethertype)

        # ipv6
        self.assertTrue(p_ipv6)
        self.assertEqual(6, p_ipv6.version)
        self.assertEqual(0, p_ipv6.traffic_class)
        self.assertEqual(0, p_ipv6.flow_label)
        self.assertEqual(len(s_buf), p_ipv6.payload_length)
        self.assertEqual(inet.IPPROTO_SCTP, p_ipv6.nxt)
        self.assertEqual(255, p_ipv6.hop_limit)
        self.assertEqual('10::10', p_ipv6.src)
        self.assertEqual('20::20', p_ipv6.dst)

        # sctp
        self.assertTrue(p_sctp)
        self.assertEqual(1, p_sctp.src_port)
        self.assertEqual(1, p_sctp.dst_port)
        self.assertEqual(0, p_sctp.vtag)
        assert isinstance(p_sctp.chunks[0], sctp.chunk_data)
        self.assertEqual(0, p_sctp.chunks[0]._type)
        self.assertEqual(0, p_sctp.chunks[0].unordered)
        self.assertEqual(0, p_sctp.chunks[0].begin)
        self.assertEqual(0, p_sctp.chunks[0].end)
        self.assertEqual(16 + len(self.payload), p_sctp.chunks[0].length)
        self.assertEqual(0, p_sctp.chunks[0].tsn)
        self.assertEqual(0, p_sctp.chunks[0].sid)
        self.assertEqual(0, p_sctp.chunks[0].seq)
        self.assertEqual(0, p_sctp.chunks[0].payload_id)
        self.assertEqual(self.payload, p_sctp.chunks[0].payload_data)
        self.assertEqual(len(s_buf), len(p_sctp))

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IPV6}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv6_values = {'version': 6,
                       'traffic_class': 0,
                       'flow_label': 0,
                       'payload_length': len(s_buf),
                       'nxt': inet.IPPROTO_SCTP,
                       'hop_limit': 255,
                       'src': '10::10',
                       'dst': '20::20',
                       'ext_hdrs': []}
        _ipv6_str = ','.join(['%s=%s' % (k, repr(ipv6_values[k]))
                              for k, v in inspect.getmembers(p_ipv6)
                              if k in ipv6_values])
        ipv6_str = '%s(%s)' % (ipv6.ipv6.__name__, _ipv6_str)

        data_values = {'unordered': 0,
                       'begin': 0,
                       'end': 0,
                       'length': 16 + len(self.payload),
                       'tsn': 0,
                       'sid': 0,
                       'seq': 0,
                       'payload_id': 0,
                       'payload_data': self.payload}
        _data_str = ','.join(['%s=%s' % (k, repr(data_values[k]))
                              for k in sorted(data_values.keys())])
        data_str = '[%s(%s)]' % (sctp.chunk_data.__name__, _data_str)

        sctp_values = {'src_port': 1,
                       'dst_port': 1,
                       'vtag': 0,
                       'csum': repr(p_sctp.csum),
                       'chunks': data_str}
        _sctp_str = ','.join(['%s=%s' % (k, sctp_values[k])
                              for k, _ in inspect.getmembers(p_sctp)
                              if k in sctp_values])
        sctp_str = '%s(%s)' % (sctp.sctp.__name__, _sctp_str)

        pkt_str = '%s, %s, %s' % (eth_str, ipv6_str, sctp_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv6_str, str(p_ipv6))
        self.assertEqual(ipv6_str, repr(p_ipv6))

        self.assertEqual(sctp_str, str(p_sctp))
        self.assertEqual(sctp_str, repr(p_sctp))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_ipv6_icmpv6(self):
        # build packet
        e = ethernet.ethernet(ethertype=ether.ETH_TYPE_IPV6)
        ip = ipv6.ipv6(nxt=inet.IPPROTO_ICMPV6)
        ic = icmpv6.icmpv6()

        p = e / ip / ic
        p.serialize()

        ipaddr = addrconv.ipv6.text_to_bin('::')

        # ethernet !6s6sH
        e_buf = b'\xff\xff\xff\xff\xff\xff' \
            + b'\x00\x00\x00\x00\x00\x00' \
            + b'\x86\xdd'

        # ipv6 !IHBB16s16s'
        ip_buf = b'\x60\x00\x00\x00' \
            + b'\x00\x00' \
            + b'\x3a' \
            + b'\xff' \
            + b'\x00\x00' \
            + ipaddr \
            + ipaddr

        # icmpv6 !BBH
        ic_buf = b'\x00' \
            + b'\x00' \
            + b'\x00\x00'

        buf = e_buf + ip_buf + ic_buf

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_ipv6 = protocols['ipv6']
        p_icmpv6 = protocols['icmpv6']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual('ff:ff:ff:ff:ff:ff', p_eth.dst)
        self.assertEqual('00:00:00:00:00:00', p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IPV6, p_eth.ethertype)

        # ipv6
        self.assertTrue(p_ipv6)
        self.assertEqual(6, p_ipv6.version)
        self.assertEqual(0, p_ipv6.traffic_class)
        self.assertEqual(0, p_ipv6.flow_label)
        self.assertEqual(len(ic_buf), p_ipv6.payload_length)
        self.assertEqual(inet.IPPROTO_ICMPV6, p_ipv6.nxt)
        self.assertEqual(255, p_ipv6.hop_limit)
        self.assertEqual('10::10', p_ipv6.src)
        self.assertEqual('20::20', p_ipv6.dst)

        # icmpv6
        self.assertTrue(p_icmpv6)
        self.assertEqual(0, p_icmpv6.type_)
        self.assertEqual(0, p_icmpv6.code)
        self.assertEqual(len(ic_buf), len(p_icmpv6))
        t = bytearray(ic_buf)
        struct.pack_into('!H', t, 2, p_icmpv6.csum)
        ph = struct.pack('!16s16sI3xB', ipaddr, ipaddr, len(ic_buf), 58)
        t = ph + t
        self.assertEqual(packet_utils.checksum(t), 0x60)

        # to string
        eth_values = {'dst': 'ff:ff:ff:ff:ff:ff',
                      'src': '00:00:00:00:00:00',
                      'ethertype': ether.ETH_TYPE_IPV6}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, _ in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ipv6_values = {'version': 6,
                       'traffic_class': 0,
                       'flow_label': 0,
                       'payload_length': len(ic_buf),
                       'nxt': inet.IPPROTO_ICMPV6,
                       'hop_limit': 255,
                       'src': '10::10',
                       'dst': '20::20',
                       'ext_hdrs': []}
        _ipv6_str = ','.join(['%s=%s' % (k, repr(ipv6_values[k]))
                              for k, _ in inspect.getmembers(p_ipv6)
                              if k in ipv6_values])
        ipv6_str = '%s(%s)' % (ipv6.ipv6.__name__, _ipv6_str)

        icmpv6_values = {'type_': 0,
                         'code': 0,
                         'csum': p_icmpv6.csum,
                         'data': b''}
        _icmpv6_str = ','.join(['%s=%s' % (k, repr(icmpv6_values[k]))
                                for k, _ in inspect.getmembers(p_icmpv6)
                                if k in icmpv6_values])
        icmpv6_str = '%s(%s)' % (icmpv6.icmpv6.__name__, _icmpv6_str)

        pkt_str = '%s, %s, %s' % (eth_str, ipv6_str, icmpv6_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(ipv6_str, str(p_ipv6))
        self.assertEqual(ipv6_str, repr(p_ipv6))

        self.assertEqual(icmpv6_str, str(p_icmpv6))
        self.assertEqual(icmpv6_str, repr(p_icmpv6))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_llc_bpdu(self):
        # buid packet
        e = ethernet.ethernet(self.dst_mac, self.src_mac,
                              ether.ETH_TYPE_IEEE802_3)
        llc_control = llc.ControlFormatU(0, 0, 0)
        l = llc.llc(llc.SAP_BPDU, llc.SAP_BPDU, llc_control)
        b = bpdu.ConfigurationBPDUs(flags=0,
                                    root_priority=32768,
                                    root_system_id_extension=0,
                                    root_mac_address=self.src_mac,
                                    root_path_cost=0,
                                    bridge_priority=32768,
                                    bridge_system_id_extension=0,
                                    bridge_mac_address=self.dst_mac,
                                    port_priority=128,
                                    port_number=4,
                                    message_age=1,
                                    max_age=20,
                                    hello_time=2,
                                    forward_delay=15)

        p = packet.Packet()
        p.add_protocol(e)
        p.add_protocol(l)
        p.add_protocol(b)
        p.serialize()

        # ethernet !6s6sH
        e_buf = self.dst_mac_bin + self.src_mac_bin + b'\x05\xdc'

        # llc !BBB
        l_buf = (b'\x42'
                 b'\x42'
                 b'\x03')

        # bpdu !HBBBQIQHHHHH
        b_buf = (b'\x00\x00'
                 b'\x00'
                 b'\x00'
                 b'\x00'
                 b'\x80\x00\xbb\xbb\xbb\xbb\xbb\xbb'
                 b'\x00\x00\x00\x00'
                 b'\x80\x00\xaa\xaa\xaa\xaa\xaa\xaa'
                 b'\x80\x04'
                 b'\x01\x00'
                 b'\x14\x00'
                 b'\x02\x00'
                 b'\x0f\x00')

        buf = e_buf + l_buf + b_buf

        # Append padding if ethernet frame is less than 60 bytes length
        pad_len = 60 - len(buf)
        if pad_len > 0:
            buf += b'\x00' * pad_len
        self.assertEqual(buf, p.data)

        # parse
        pkt = packet.Packet(p.data)
        protocols = self.get_protocols(pkt)
        p_eth = protocols['ethernet']
        p_llc = protocols['llc']
        p_bpdu = protocols['ConfigurationBPDUs']

        # ethernet
        self.assertTrue(p_eth)
        self.assertEqual(self.dst_mac, p_eth.dst)
        self.assertEqual(self.src_mac, p_eth.src)
        self.assertEqual(ether.ETH_TYPE_IEEE802_3, p_eth.ethertype)

        # llc
        self.assertTrue(p_llc)
        self.assertEqual(llc.SAP_BPDU, p_llc.dsap_addr)
        self.assertEqual(llc.SAP_BPDU, p_llc.ssap_addr)
        self.assertEqual(0, p_llc.control.modifier_function1)
        self.assertEqual(0, p_llc.control.pf_bit)
        self.assertEqual(0, p_llc.control.modifier_function2)

        # bpdu
        self.assertTrue(p_bpdu)
        self.assertEqual(bpdu.PROTOCOL_IDENTIFIER, p_bpdu._protocol_id)
        self.assertEqual(bpdu.PROTOCOLVERSION_ID_BPDU, p_bpdu._version_id)
        self.assertEqual(bpdu.TYPE_CONFIG_BPDU, p_bpdu._bpdu_type)
        self.assertEqual(0, p_bpdu.flags)
        self.assertEqual(32768, p_bpdu.root_priority)
        self.assertEqual(0, p_bpdu.root_system_id_extension)
        self.assertEqual(self.src_mac, p_bpdu.root_mac_address)
        self.assertEqual(0, p_bpdu.root_path_cost)
        self.assertEqual(32768, p_bpdu.bridge_priority)
        self.assertEqual(0, p_bpdu.bridge_system_id_extension)
        self.assertEqual(self.dst_mac, p_bpdu.bridge_mac_address)
        self.assertEqual(128, p_bpdu.port_priority)
        self.assertEqual(4, p_bpdu.port_number)
        self.assertEqual(1, p_bpdu.message_age)
        self.assertEqual(20, p_bpdu.max_age)
        self.assertEqual(2, p_bpdu.hello_time)
        self.assertEqual(15, p_bpdu.forward_delay)

        # to string
        eth_values = {'dst': self.dst_mac,
                      'src': self.src_mac,
                      'ethertype': ether.ETH_TYPE_IEEE802_3}
        _eth_str = ','.join(['%s=%s' % (k, repr(eth_values[k]))
                             for k, v in inspect.getmembers(p_eth)
                             if k in eth_values])
        eth_str = '%s(%s)' % (ethernet.ethernet.__name__, _eth_str)

        ctrl_values = {'modifier_function1': 0,
                       'pf_bit': 0,
                       'modifier_function2': 0}
        _ctrl_str = ','.join(['%s=%s' % (k, repr(ctrl_values[k]))
                              for k, v in inspect.getmembers(p_llc.control)
                              if k in ctrl_values])
        ctrl_str = '%s(%s)' % (llc.ControlFormatU.__name__, _ctrl_str)

        llc_values = {'dsap_addr': repr(llc.SAP_BPDU),
                      'ssap_addr': repr(llc.SAP_BPDU),
                      'control': ctrl_str}
        _llc_str = ','.join(['%s=%s' % (k, llc_values[k])
                             for k, v in inspect.getmembers(p_llc)
                             if k in llc_values])
        llc_str = '%s(%s)' % (llc.llc.__name__, _llc_str)

        bpdu_values = {'flags': 0,
                       'root_priority': int(32768),
                       'root_system_id_extension': int(0),
                       'root_mac_address': self.src_mac,
                       'root_path_cost': 0,
                       'bridge_priority': int(32768),
                       'bridge_system_id_extension': int(0),
                       'bridge_mac_address': self.dst_mac,
                       'port_priority': 128,
                       'port_number': 4,
                       'message_age': float(1),
                       'max_age': float(20),
                       'hello_time': float(2),
                       'forward_delay': float(15)}
        _bpdu_str = ','.join(['%s=%s' % (k, repr(bpdu_values[k]))
                              for k, v in inspect.getmembers(p_bpdu)
                              if k in bpdu_values])
        bpdu_str = '%s(%s)' % (bpdu.ConfigurationBPDUs.__name__, _bpdu_str)

        pkt_str = '%s, %s, %s' % (eth_str, llc_str, bpdu_str)

        self.assertEqual(eth_str, str(p_eth))
        self.assertEqual(eth_str, repr(p_eth))

        self.assertEqual(llc_str, str(p_llc))
        self.assertEqual(llc_str, repr(p_llc))

        self.assertEqual(bpdu_str, str(p_bpdu))
        self.assertEqual(bpdu_str, repr(p_bpdu))

        self.assertEqual(pkt_str, str(pkt))
        self.assertEqual(pkt_str, repr(pkt))

    def test_div_api(self):
        e = ethernet.ethernet(self.dst_mac, self.src_mac, ether.ETH_TYPE_IP)
        i = ipv4.ipv4()
        u = udp.udp(self.src_port, self.dst_port)
        pkt = e / i / u
        self.assertTrue(isinstance(pkt, packet.Packet))
        self.assertTrue(isinstance(pkt.protocols[0], ethernet.ethernet))
        self.assertTrue(isinstance(pkt.protocols[1], ipv4.ipv4))
        self.assertTrue(isinstance(pkt.protocols[2], udp.udp))



#!/usr/bin/env python
# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from subprocess import Popen, PIPE, STDOUT
import time

from mininet.net import Mininet
from mininet.node import RemoteController, OVSKernelSwitch

TIMEOUT = 60
OSKEN_HOST = '127.0.0.1'
OSKEN_PORT = 6633
PYTHON_BIN = '.venv/bin/python'
OSKEN_MGR = 'osken-manager'


class OVS12KernelSwitch(OVSKernelSwitch):
    """Set protocols parameter for OVS version 1.10"""

    def start(self, controllers):
        super(OVS12KernelSwitch, self).start(controllers)
        self.cmd('ovs-vsctl set Bridge', self,
                 "protocols='[OpenFlow10, OpenFlow12]'")


class TestWithOVS12(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mn = Mininet()
        c = cls.mn.addController(controller=RemoteController,
                                 ip=OSKEN_HOST, port=OSKEN_PORT)
        c.start()

        s1 = cls.mn.addSwitch('s1', cls=OVS12KernelSwitch)
        s1.start(cls.mn.controllers)

        h1 = cls.mn.addHost('h1', ip='0.0.0.0/0')

        link = cls.mn.addLink(h1, s1)
        s1.attach(link.intf2)

    @classmethod
    def tearDownClass(cls):
        cls.mn.stop()

    def test_add_flow_v10(self):
        app = 'os_ken/tests/integrated/test_add_flow_v10.py'
        self._run_os_ken_manager_and_check_output(app)

    def test_request_reply_v12(self):
        app = 'os_ken/tests/integrated/test_request_reply_v12.py'
        self._run_os_ken_manager_and_check_output(app)

    def test_add_flow_v12_actions(self):
        app = 'os_ken/tests/integrated/test_add_flow_v12_actions.py'
        self._run_os_ken_manager_and_check_output(app)

    def test_add_flow_v12_matches(self):
        app = 'os_ken/tests/integrated/test_add_flow_v12_matches.py'
        self._run_os_ken_manager_and_check_output(app)

    def test_of_config(self):
        self.skipTest('OVS 1.10 does not support of_config')

    def _run_os_ken_manager_and_check_output(self, app):
        cmd = [PYTHON_BIN, OSKEN_MGR, app]
        p = Popen(cmd, stdout=PIPE, stderr=STDOUT)

        while True:
            if p.poll() is not None:
                raise Exception('Another osken-manager already running?')

            line = p.stdout.readline().strip()
            if line == '':
                time.sleep(1)
                continue

            print("osken-manager: %s" % line)
            if line.find('TEST_FINISHED') != -1:
                self.assertTrue(line.find('Completed=[True]') != -1)
                p.terminate()
                p.communicate()  # wait for subprocess is terminated
                break



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2014 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# client for os_ken.app.ofctl.service

from os_ken.base import app_manager
from . import event


def get_datapath(app, dpid=None):
    """
    Get datapath object by dpid.

    :param app: Client OSKenApp instance
    :param dpid: Datapath ID (int type) or None to get all datapath objects

    Returns a object of datapath, a list of datapath objects when no dpid
    given or None when error.

    Raises an exception if any of the given values is invalid.

    Example::

        # ...(snip)...
        import os_ken.app.ofctl.api as ofctl_api


        class MyApp(app_manager.OSKenApp):

            def _my_handler(self, ev):
                # Get all datapath objects
                result = ofctl_api.get_datapath(self)

                # Get the datapath object which has the given dpid
                result = ofctl_api.get_datapath(self, dpid=1)
    """
    return app.send_request(event.GetDatapathRequest(dpid=dpid))()


def send_msg(app, msg, reply_cls=None, reply_multi=False):
    """
    Send an OpenFlow message and wait for reply messages.

    :param app: Client OSKenApp instance
    :param msg: An OpenFlow controller-to-switch message to send
    :param reply_cls: OpenFlow message class for expected replies.
        None means no replies are expected.  The default is None.
    :param reply_multi: True if multipart replies are expected.
        The default is False.

    If no replies, returns None.
    If reply_multi=False, returns OpenFlow switch-to-controller message.
    If reply_multi=True, returns a list of OpenFlow switch-to-controller
    messages.

    Raise an exception on error.

    Example::

        # ...(snip)...
        import os_ken.app.ofctl.api as ofctl_api


        class MyApp(app_manager.OSKenApp):

            def _my_handler(self, ev):
                # ...(snip)...
                msg = parser.OFPPortDescStatsRequest(datapath=datapath)
                result = ofctl_api.send_msg(
                    self, msg,
                    reply_cls=parser.OFPPortDescStatsReply,
                    reply_multi=True)
    """
    return app.send_request(event.SendMsgRequest(msg=msg,
                                                 reply_cls=reply_cls,
                                                 reply_multi=reply_multi))()


app_manager.require_app('os_ken.app.ofctl.service', api_style=True)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on netconf.xsd

# rpc
RPC = 'rpc'
MESSAGE_ID = 'message-id'      # message-id attribute

# error
TRANSPORT = 'transport'
PROTOCOL = 'protocol'
APPLICATION = 'application'

# error-tag
IN_USE = 'in-use'
INVALID_VALUE = 'invalid-value'
TOO_BIG = 'too-big'
MISSING_ATTRIBUTE = 'missing-attribute'
BAD_ATTRIBUTE = 'bad-attribute'
UNKNOWN_ATTRIBUTE = 'unknown-attribute'
MISSING_ELEMENT = 'missing-element'
BAD_ELEMENT = 'bad-element'
UNKNOWN_ELEMENT = 'unknown-element'
UNKNOWN_NAMESPACE = 'unknown-namespace'
ACCESS_DENIED = 'access-denied'
LOCK_DENIED = 'lock-denied'
RESOURCE_DENIED = 'resource-denied'
ROLLBACK_FAILED = 'rollback-failed'
DATA_EXISTS = 'data-exists'
DATA_MISSING = 'data-missing'
OPERATION_NOT_SUPPORTED = 'operation-not-supported'
OPERATION_FAILED = 'operation-failed'
PARTIAL_OPERATION = 'partial-operation'
MALFORMED_MESSAGE = 'malformed-message'

# error-severity
ERROR = 'error'
WARNING = 'warning'

# error-info
# bad-attribute, bad-element and ok-element are defined above
# BAD_ATTRIBUTE = 'bad-attribute'
# BAD_ELEMENT = 'bad-element'
# OK_ELEMENT = 'ok-element'
ERR_ELEMENT = 'err-element'
NOOP_ELEMENT = 'noop-element'
BAD_NAMESPACE = 'bad-namespace'

# rpc-error
ERROR_TYPE = 'error-type'
ERROR_TAG = 'error-tag'
ERROR_SEVERITY = 'error-severity'
ERROR_APP_TAG = 'error-app-tag'
ERROR_PATH = 'error-path'
ERROR_MESSAGE = 'error-message'
ERROR_INFO = 'error-info'

# edit-operation
OPERATION = 'operation'         # operation attribute
MERGE = 'merge'
REPLACE = 'replace'
CREATE = 'create'
DELETE = 'delete'
REMOVE = 'remove'

# default-operation
# merge and replace are defined above
# MERGE = 'merge'
# REPLACE = 'replace'
NONE = 'none'
DEFAULT_OPERATION = 'default-operation'

# rpc-reply
OK = 'ok'
RPC_REPLY = 'rpc-reply'

# data-inline
DATA = 'data'
RPC_ERROR = 'rpc-error'

# rpc-operation
RPCOPERATION = 'rpcOperation'

# rpc-response
RPCRESPONSE = 'rpcResponse'
HELLO = 'hello'
CAPABILITIES = 'capabilities'
CAPABILITY = 'capability'

# config-inline
CONFIG = 'config'

# config-name
CONFIG_NAME = 'config-name'
STARTUP = 'startup'
CANDIDATE = 'candidate'
RUNNING = 'running'

# config-uri
URL = 'url'

# rpc-operation-source
SOURCE = 'source'

# rpc-operation-target
TARGET = 'target'

# filter
SUBTREE = 'subtree'
XPATH = 'xpath'

# filter-inline
TYPE = 'type'           # type attribute
FILTER = 'filter'

# test-option
TEST_THEN_SET = 'test-then-set'
SET = 'set'
TEST_OPTION = 'test-option'

# error-option
STOP_ON_ERROR = 'stop-on-error'
IGNORE_ERROR = 'ignore-error'
ROLLBACK_ON_ERROR = 'rollback-on-error'
ERROR_OPTION = 'error-option'

# get
GET = 'get'

# get-config
GET_CONFIG = 'get-config'

# edit-config
EDIT_CONFIG = 'edit-config'

# copy-config
COPY_CONFIG = 'copy-config'

# delete-config
DELETE_CONFIG = 'delete-config'

# lock
LOCK = 'lock'

# unlock
UNLOCK = 'unlock'

# validate
VALIDATE = 'validate'

# commit
CONFIRMED = 'confirmed'
CONFIRM_TIMEOUT = 'confirm-timeout'
PERSIST = 'persist'
PERSIST_ID = 'persist-id'
COMMIT = 'commit'

# cancel-commit
# persist-id is defined above
# PERSIST_ID = 'persist-id'
CANCEL_COMMIT = 'cancel-commit'

# discard-changes
DISCARD_CHANGES = 'discard-changes'

# close-session
CLOSE_SESSION = 'close-session'

# kill-session
KILL_SESSION = 'kill-session'



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2014 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_ken.ofproto import ofproto_v1_0
from os_ken.ofproto import ofproto_v1_0_parser
from os_ken.ofproto import ofproto_v1_2
from os_ken.ofproto import ofproto_v1_2_parser
from os_ken.ofproto import ofproto_v1_3
from os_ken.ofproto import ofproto_v1_3_parser
from os_ken.ofproto import ofproto_v1_4
from os_ken.ofproto import ofproto_v1_4_parser
from os_ken.ofproto import ofproto_v1_5
from os_ken.ofproto import ofproto_v1_5_parser


_versions = {
    ofproto_v1_0.OFP_VERSION: (ofproto_v1_0, ofproto_v1_0_parser),
    ofproto_v1_2.OFP_VERSION: (ofproto_v1_2, ofproto_v1_2_parser),
    ofproto_v1_3.OFP_VERSION: (ofproto_v1_3, ofproto_v1_3_parser),
    ofproto_v1_4.OFP_VERSION: (ofproto_v1_4, ofproto_v1_4_parser),
    ofproto_v1_5.OFP_VERSION: (ofproto_v1_5, ofproto_v1_5_parser),
}


# OF versions supported by every apps in this process (intersection)
_supported_versions = set(_versions.keys())


def set_app_supported_versions(vers):
    global _supported_versions

    _supported_versions &= set(vers)
    assert _supported_versions, 'No OpenFlow version is available'


class ProtocolDesc(object):
    """
    OpenFlow protocol version flavor descriptor
    """

    def __init__(self, version=None):
        if version is None:
            version = max(_supported_versions)
        self.set_version(version)

    def set_version(self, version):
        assert version in _supported_versions
        (self.ofproto, self.ofproto_parser) = _versions[version]

    @property
    def supported_ofp_version(self):
        return _supported_versions



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Running or runtime configuration base classes.
"""
import abc
from abc import ABCMeta
from abc import abstractmethod
import functools
import numbers
import logging
import uuid

from os_ken.services.protocols.bgp.base import add_bgp_error_metadata
from os_ken.services.protocols.bgp.base import BGPSException
from os_ken.services.protocols.bgp.base import get_validator
from os_ken.services.protocols.bgp.base import RUNTIME_CONF_ERROR_CODE
from os_ken.services.protocols.bgp.base import validate
from os_ken.services.protocols.bgp.utils import validation
from os_ken.services.protocols.bgp.utils.validation import is_valid_asn

LOG = logging.getLogger('bgpspeaker.rtconf.base')

#
# Nested settings.
#
CAP_REFRESH = 'cap_refresh'
CAP_ENHANCED_REFRESH = 'cap_enhanced_refresh'
CAP_FOUR_OCTET_AS_NUMBER = 'cap_four_octet_as_number'
CAP_MBGP_IPV4 = 'cap_mbgp_ipv4'
CAP_MBGP_IPV6 = 'cap_mbgp_ipv6'
CAP_MBGP_VPNV4 = 'cap_mbgp_vpnv4'
CAP_MBGP_VPNV6 = 'cap_mbgp_vpnv6'
CAP_MBGP_EVPN = 'cap_mbgp_evpn'
CAP_MBGP_IPV4FS = 'cap_mbgp_ipv4fs'
CAP_MBGP_IPV6FS = 'cap_mbgp_ipv6fs'
CAP_MBGP_VPNV4FS = 'cap_mbgp_vpnv4fs'
CAP_MBGP_VPNV6FS = 'cap_mbgp_vpnv6fs'
CAP_MBGP_L2VPNFS = 'cap_mbgp_l2vpnfs'
CAP_RTC = 'cap_rtc'
RTC_AS = 'rtc_as'
HOLD_TIME = 'hold_time'

# To control how many prefixes can be received from a neighbor.
# 0 value indicates no limit and other related options will be ignored.
# Current behavior is to log that limit has reached.
MAX_PREFIXES = 'max_prefixes'

# Has same meaning as: http://www.juniper.net/techpubs/software/junos/junos94
# /swconfig-routing/disabling-suppression-of-route-
# advertisements.html#id-13255463
ADVERTISE_PEER_AS = 'advertise_peer_as'

# MED - MULTI_EXIT_DISC
MULTI_EXIT_DISC = 'multi_exit_disc'

# Extended community attribute route origin.
SITE_OF_ORIGINS = 'site_of_origins'

# Constants related to errors.
CONF_NAME = 'conf_name'
CONF_VALUE = 'conf_value'

# Max. value  limits
MAX_NUM_IMPORT_RT = 1000
MAX_NUM_EXPORT_RT = 250
MAX_NUM_SOO = 10


# =============================================================================
# Runtime configuration errors or exceptions.
# =============================================================================

@add_bgp_error_metadata(code=RUNTIME_CONF_ERROR_CODE, sub_code=1,
                        def_desc='Error with runtime-configuration.')
class RuntimeConfigError(BGPSException):
    """Base class for all runtime configuration errors.
    """
    pass


@add_bgp_error_metadata(code=RUNTIME_CONF_ERROR_CODE, sub_code=2,
                        def_desc='Missing required configuration.')
class MissingRequiredConf(RuntimeConfigError):
    """Exception raised when trying to configure with missing required
    settings.
    """

    def __init__(self, **kwargs):
        conf_name = kwargs.get('conf_name')
        if conf_name:
            super(MissingRequiredConf, self).__init__(
                desc='Missing required configuration: %s' % conf_name)
        else:
            super(MissingRequiredConf, self).__init__(desc=kwargs.get('desc'))


@add_bgp_error_metadata(code=RUNTIME_CONF_ERROR_CODE, sub_code=3,
                        def_desc='Incorrect Type for configuration.')
class ConfigTypeError(RuntimeConfigError):
    """Exception raised when configuration value type miss-match happens.
    """

    def __init__(self, **kwargs):
        conf_name = kwargs.get(CONF_NAME)
        conf_value = kwargs.get(CONF_VALUE)
        if conf_name and conf_value:
            super(ConfigTypeError, self).__init__(
                desc='Incorrect Type %s for configuration: %s' %
                (conf_value, conf_name))
        elif conf_name:
            super(ConfigTypeError, self).__init__(
                desc='Incorrect Type for configuration: %s' % conf_name)
        else:
            super(ConfigTypeError, self).__init__(desc=kwargs.get('desc'))


@add_bgp_error_metadata(code=RUNTIME_CONF_ERROR_CODE, sub_code=4,
                        def_desc='Incorrect Value for configuration.')
class ConfigValueError(RuntimeConfigError):
    """Exception raised when configuration value is of correct type but
    incorrect value.
    """

    def __init__(self, **kwargs):
        conf_name = kwargs.get(CONF_NAME)
        conf_value = kwargs.get(CONF_VALUE)
        if conf_name and conf_value:
            super(ConfigValueError, self).__init__(
                desc='Incorrect Value %s for configuration: %s' %
                (conf_value, conf_name))
        elif conf_name:
            super(ConfigValueError, self).__init__(
                desc='Incorrect Value for configuration: %s' % conf_name)
        else:
            super(ConfigValueError, self).__init__(desc=kwargs.get('desc'))


# =============================================================================
# Configuration base classes.
# =============================================================================

class BaseConf(object, metaclass=abc.ABCMeta):
    """Base class for a set of configuration values.

    Configurations can be required or optional. Also acts as a container of
    configuration change listeners.
    """

    def __init__(self, **kwargs):
        self._req_settings = self.get_req_settings()
        self._opt_settings = self.get_opt_settings()
        self._valid_evts = self.get_valid_evts()
        self._listeners = {}
        self._settings = {}

        # validate required and unknown settings
        self._validate_req_unknown_settings(**kwargs)

        # Initialize configuration settings.
        self._init_req_settings(**kwargs)
        self._init_opt_settings(**kwargs)

    @property
    def settings(self):
        """Returns a copy of current settings."""
        return self._settings.copy()

    @classmethod
    def get_valid_evts(cls):
        return set()

    @classmethod
    def get_req_settings(cls):
        return set()

    @classmethod
    def get_opt_settings(cls):
        return set()

    @abstractmethod
    def _init_opt_settings(self, **kwargs):
        """Sub-classes should override this method to initialize optional
         settings.
        """
        pass

    @abstractmethod
    def update(self, **kwargs):
        # Validate given values
        self._validate_req_unknown_settings(**kwargs)

    def _validate_req_unknown_settings(self, **kwargs):
        """Checks if required settings are present.

        Also checks if unknown requirements are present.
        """
        # Validate given configuration.
        self._all_attrs = (self._req_settings | self._opt_settings)
        if not kwargs and len(self._req_settings) > 0:
            raise MissingRequiredConf(desc='Missing all required attributes.')

        given_attrs = frozenset(kwargs.keys())
        unknown_attrs = given_attrs - self._all_attrs
        if unknown_attrs:
            raise RuntimeConfigError(desc=(
                'Unknown attributes: %s' %
                ', '.join([str(i) for i in unknown_attrs])))
        missing_req_settings = self._req_settings - given_attrs
        if missing_req_settings:
            raise MissingRequiredConf(conf_name=list(missing_req_settings))

    def _init_req_settings(self, **kwargs):
        for req_attr in self._req_settings:
            req_attr_value = kwargs.get(req_attr)
            if req_attr_value is None:
                raise MissingRequiredConf(conf_name=req_attr_value)
            # Validate attribute value
            req_attr_value = get_validator(req_attr)(req_attr_value)
            self._settings[req_attr] = req_attr_value

    def add_listener(self, evt, callback):
        #   if (evt not in self.get_valid_evts()):
        #       raise RuntimeConfigError(desc=('Unknown event %s' % evt))

        listeners = self._listeners.get(evt, None)
        if not listeners:
            listeners = set()
            self._listeners[evt] = listeners
        listeners.update([callback])

    def remove_listener(self, evt, callback):
        if evt in self.get_valid_evts():
            listeners = self._listeners.get(evt, None)
            if listeners and (callback in listeners):
                listeners.remove(callback)
                return True

        return False

    def _notify_listeners(self, evt, value):
        listeners = self._listeners.get(evt, [])
        for callback in listeners:
            callback(ConfEvent(self, evt, value))

    def __repr__(self):
        return '%s(%r)' % (self.__class__, self._settings)


class ConfWithId(BaseConf):
    """Configuration settings related to identity."""
    # Config./resource identifier.
    ID = 'id'
    # Config./resource name.
    NAME = 'name'
    # Config./resource description.
    DESCRIPTION = 'description'

    UPDATE_NAME_EVT = 'update_name_evt'
    UPDATE_DESCRIPTION_EVT = 'update_description_evt'

    VALID_EVT = frozenset([UPDATE_NAME_EVT, UPDATE_DESCRIPTION_EVT])
    OPTIONAL_SETTINGS = frozenset([ID, NAME, DESCRIPTION])

    def __init__(self, **kwargs):
        super(ConfWithId, self).__init__(**kwargs)

    @classmethod
    def get_opt_settings(cls):
        self_confs = super(ConfWithId, cls).get_opt_settings()
        self_confs.update(ConfWithId.OPTIONAL_SETTINGS)
        return self_confs

    @classmethod
    def get_req_settings(cls):
        self_confs = super(ConfWithId, cls).get_req_settings()
        return self_confs

    @classmethod
    def get_valid_evts(cls):
        self_valid_evts = super(ConfWithId, cls).get_valid_evts()
        self_valid_evts.update(ConfWithId.VALID_EVT)
        return self_valid_evts

    def _init_opt_settings(self, **kwargs):
        super(ConfWithId, self)._init_opt_settings(**kwargs)
        self._settings[ConfWithId.ID] = \
            compute_optional_conf(ConfWithId.ID, str(uuid.uuid4()), **kwargs)
        self._settings[ConfWithId.NAME] = \
            compute_optional_conf(ConfWithId.NAME, str(self), **kwargs)
        self._settings[ConfWithId.DESCRIPTION] = \
            compute_optional_conf(ConfWithId.DESCRIPTION, str(self), **kwargs)

    @property
    def id(self):
        return self._settings[ConfWithId.ID]

    @property
    def name(self):
        return self._settings[ConfWithId.NAME]

    @name.setter
    def name(self, new_name):
        old_name = self.name
        if not new_name:
            new_name = repr(self)
        else:
            get_validator(ConfWithId.NAME)(new_name)

        if old_name != new_name:
            self._settings[ConfWithId.NAME] = new_name
            self._notify_listeners(ConfWithId.UPDATE_NAME_EVT,
                                   (old_name, self.name))

    @property
    def description(self):
        return self._settings[ConfWithId.DESCRIPTION]

    @description.setter
    def description(self, new_description):
        old_desc = self.description
        if not new_description:
            new_description = str(self)
        else:
            get_validator(ConfWithId.DESCRIPTION)(new_description)

        if old_desc != new_description:
            self._settings[ConfWithId.DESCRIPTION] = new_description
            self._notify_listeners(ConfWithId.UPDATE_DESCRIPTION_EVT,
                                   (old_desc, self.description))

    def update(self, **kwargs):
        # Update inherited configurations
        super(ConfWithId, self).update(**kwargs)
        self.name = compute_optional_conf(ConfWithId.NAME,
                                          str(self),
                                          **kwargs)
        self.description = compute_optional_conf(ConfWithId.DESCRIPTION,
                                                 str(self),
                                                 **kwargs)


class ConfWithStats(BaseConf):
    """Configuration settings related to statistics collection."""

    # Enable or disable statistics logging.
    STATS_LOG_ENABLED = 'statistics_log_enabled'
    DEFAULT_STATS_LOG_ENABLED = False

    # Statistics logging time.
    STATS_TIME = 'statistics_interval'
    DEFAULT_STATS_TIME = 60

    UPDATE_STATS_LOG_ENABLED_EVT = 'update_stats_log_enabled_evt'
    UPDATE_STATS_TIME_EVT = 'update_stats_time_evt'

    VALID_EVT = frozenset([UPDATE_STATS_LOG_ENABLED_EVT,
                           UPDATE_STATS_TIME_EVT])
    OPTIONAL_SETTINGS = frozenset([STATS_LOG_ENABLED, STATS_TIME])

    def __init__(self, **kwargs):
        super(ConfWithStats, self).__init__(**kwargs)

    def _init_opt_settings(self, **kwargs):
        super(ConfWithStats, self)._init_opt_settings(**kwargs)
        self._settings[ConfWithStats.STATS_LOG_ENABLED] = \
            compute_optional_conf(ConfWithStats.STATS_LOG_ENABLED,
                                  ConfWithStats.DEFAULT_STATS_LOG_ENABLED,
                                  **kwargs)
        self._settings[ConfWithStats.STATS_TIME] = \
            compute_optional_conf(ConfWithStats.STATS_TIME,
                                  ConfWithStats.DEFAULT_STATS_TIME,
                                  **kwargs)

    @property
    def stats_log_enabled(self):
        return self._settings[ConfWithStats.STATS_LOG_ENABLED]

    @stats_log_enabled.setter
    def stats_log_enabled(self, enabled):
        get_validator(ConfWithStats.STATS_LOG_ENABLED)(enabled)
        if enabled != self.stats_log_enabled:
            self._settings[ConfWithStats.STATS_LOG_ENABLED] = enabled
            self._notify_listeners(ConfWithStats.UPDATE_STATS_LOG_ENABLED_EVT,
                                   enabled)

    @property
    def stats_time(self):
        return self._settings[ConfWithStats.STATS_TIME]

    @stats_time.setter
    def stats_time(self, stats_time):
        get_validator(ConfWithStats.STATS_TIME)(stats_time)
        if stats_time != self.stats_time:
            self._settings[ConfWithStats.STATS_TIME] = stats_time
            self._notify_listeners(ConfWithStats.UPDATE_STATS_TIME_EVT,
                                   stats_time)

    @classmethod
    def get_opt_settings(cls):
        confs = super(ConfWithStats, cls).get_opt_settings()
        confs.update(ConfWithStats.OPTIONAL_SETTINGS)
        return confs

    @classmethod
    def get_valid_evts(cls):
        valid_evts = super(ConfWithStats, cls).get_valid_evts()
        valid_evts.update(ConfWithStats.VALID_EVT)
        return valid_evts

    def update(self, **kwargs):
        # Update inherited configurations
        super(ConfWithStats, self).update(**kwargs)
        self.stats_log_enabled = \
            compute_optional_conf(ConfWithStats.STATS_LOG_ENABLED,
                                  ConfWithStats.DEFAULT_STATS_LOG_ENABLED,
                                  **kwargs)
        self.stats_time = \
            compute_optional_conf(ConfWithStats.STATS_TIME,
                                  ConfWithStats.DEFAULT_STATS_TIME,
                                  **kwargs)


class BaseConfListener(object, metaclass=abc.ABCMeta):
    """Base class of all configuration listeners."""

    def __init__(self, base_conf):
        pass
    # TODO(PH): re-vist later and check if we need this check
#         if not isinstance(base_conf, BaseConf):
#             raise TypeError('Currently we only support listening to '
#                             'instances of BaseConf')


class ConfWithIdListener(BaseConfListener):

    def __init__(self, conf_with_id):
        assert conf_with_id
        super(ConfWithIdListener, self).__init__(conf_with_id)
        conf_with_id.add_listener(ConfWithId.UPDATE_NAME_EVT,
                                  self.on_chg_name_conf_with_id)
        conf_with_id.add_listener(ConfWithId.UPDATE_DESCRIPTION_EVT,
                                  self.on_chg_desc_conf_with_id)

    def on_chg_name_conf_with_id(self, conf_evt):
        # Note did not makes this method abstract as this is not important
        # event.
        raise NotImplementedError()

    def on_chg_desc_conf_with_id(self, conf_evt):
        # Note did not makes this method abstract as this is not important
        # event.
        raise NotImplementedError()


class ConfWithStatsListener(BaseConfListener):

    def __init__(self, conf_with_stats):
        assert conf_with_stats
        super(ConfWithStatsListener, self).__init__(conf_with_stats)
        conf_with_stats.add_listener(
            ConfWithStats.UPDATE_STATS_LOG_ENABLED_EVT,
            self.on_chg_stats_enabled_conf_with_stats)

        conf_with_stats.add_listener(ConfWithStats.UPDATE_STATS_TIME_EVT,
                                     self.on_chg_stats_time_conf_with_stats)

    @abstractmethod
    def on_chg_stats_time_conf_with_stats(self, conf_evt):
        raise NotImplementedError()

    @abstractmethod
    def on_chg_stats_enabled_conf_with_stats(self, conf_evt):
        raise NotImplementedError()


@functools.total_ordering
class ConfEvent(object):
    """Encapsulates configuration settings change/update event."""

    def __init__(self, evt_src, evt_name, evt_value):
        """Creates an instance using given parameters.

        Parameters:
            -`evt_src`: (BaseConf) source of the event
            -`evt_name`: (str) name of event, has to be one of the valid
            event of `evt_src`
            - `evt_value`: (tuple) event context that helps event handler
        """
        if evt_name not in evt_src.get_valid_evts():
            raise ValueError('Event %s is not a valid event for type %s.' %
                             (evt_name, type(evt_src)))
        self._src = evt_src
        self._name = evt_name
        self._value = evt_value

    @property
    def src(self):
        return self._src

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '<ConfEvent(%s, %s, %s)>' % (self.src, self.name, self.value)

    def __str__(self):
        return ('ConfEvent(src=%s, name=%s, value=%s)' %
                (self.src, self.name, self.value))

    def __lt__(self, other):
        return ((self.src, self.name, self.value) <
                (other.src, other.name, other.value))

    def __eq__(self, other):
        return ((self.src, self.name, self.value) ==
                (other.src, other.name, other.value))


# =============================================================================
# Runtime configuration setting validators and their registry.
# =============================================================================

@validate(name=ConfWithId.ID)
def validate_conf_id(identifier):
    if not isinstance(identifier, str):
        raise ConfigTypeError(conf_name=ConfWithId.ID, conf_value=identifier)
    if len(identifier) > 128:
        raise ConfigValueError(conf_name=ConfWithId.ID, conf_value=identifier)
    return identifier


@validate(name=ConfWithId.NAME)
def validate_conf_name(name):
    if not isinstance(name, str):
        raise ConfigTypeError(conf_name=ConfWithId.NAME, conf_value=name)
    if len(name) > 128:
        raise ConfigValueError(conf_name=ConfWithId.NAME, conf_value=name)
    return name


@validate(name=ConfWithId.DESCRIPTION)
def validate_conf_desc(description):
    if not isinstance(description, str):
        raise ConfigTypeError(conf_name=ConfWithId.DESCRIPTION,
                              conf_value=description)
    return description


@validate(name=ConfWithStats.STATS_LOG_ENABLED)
def validate_stats_log_enabled(stats_log_enabled):
    if not isinstance(stats_log_enabled, bool):
        raise ConfigTypeError(desc='Statistics log enabled settings can only'
                              ' be boolean type.')
    return stats_log_enabled


@validate(name=ConfWithStats.STATS_TIME)
def validate_stats_time(stats_time):
    if not isinstance(stats_time, numbers.Integral):
        raise ConfigTypeError(desc='Statistics log timer value has to be of '
                              'integral type but got: %r' % stats_time)
    if stats_time < 10:
        raise ConfigValueError(desc='Statistics log timer cannot be set to '
                               'less then 10 sec, given timer value %s.' %
                               stats_time)
    return stats_time


@validate(name=CAP_REFRESH)
def validate_cap_refresh(crefresh):
    if not isinstance(crefresh, bool):
        raise ConfigTypeError(desc='Invalid Refresh capability settings: %s. '
                              'Boolean value expected' % crefresh)
    return crefresh


@validate(name=CAP_ENHANCED_REFRESH)
def validate_cap_enhanced_refresh(cer):
    if not isinstance(cer, bool):
        raise ConfigTypeError(desc='Invalid Enhanced Refresh capability '
                              'settings: %s. Boolean value expected' % cer)
    return cer


@validate(name=CAP_FOUR_OCTET_AS_NUMBER)
def validate_cap_four_octet_as_number(cfoan):
    if not isinstance(cfoan, bool):
        raise ConfigTypeError(desc='Invalid Four-Octet AS Number capability '
                              'settings: %s boolean value expected' % cfoan)
    return cfoan


@validate(name=CAP_MBGP_IPV4)
def validate_cap_mbgp_ipv4(cmv4):
    if not isinstance(cmv4, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP IPv4 capability '
                              'settings: %s. Boolean value expected' % cmv4)

    return cmv4


@validate(name=CAP_MBGP_IPV6)
def validate_cap_mbgp_ipv6(cmv6):
    if not isinstance(cmv6, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP IPv6 capability '
                              'settings: %s. Boolean value expected' % cmv6)

    return cmv6


@validate(name=CAP_MBGP_VPNV4)
def validate_cap_mbgp_vpnv4(cmv4):
    if not isinstance(cmv4, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP VPNv4 capability '
                              'settings: %s. Boolean value expected' % cmv4)

    return cmv4


@validate(name=CAP_MBGP_VPNV6)
def validate_cap_mbgp_vpnv6(cmv6):
    if not isinstance(cmv6, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP VPNv6 capability '
                              'settings: %s. Boolean value expected' % cmv6)

    return cmv6


@validate(name=CAP_MBGP_EVPN)
def validate_cap_mbgp_evpn(cmevpn):
    if not isinstance(cmevpn, bool):
        raise ConfigTypeError(desc='Invalid Ethernet VPN capability '
                              'settings: %s. Boolean value expected' % cmevpn)
    return cmevpn


@validate(name=CAP_MBGP_IPV4FS)
def validate_cap_mbgp_ipv4fs(cmv4fs):
    if not isinstance(cmv4fs, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP '
                              'IPv4 Flow Specification capability '
                              'settings: %s. Boolean value expected' % cmv4fs)
    return cmv4fs


@validate(name=CAP_MBGP_IPV6FS)
def validate_cap_mbgp_ipv6fs(cmv6fs):
    if not isinstance(cmv6fs, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP '
                              'IPv6 Flow Specification capability '
                              'settings: %s. Boolean value expected' % cmv6fs)
    return cmv6fs


@validate(name=CAP_MBGP_VPNV4FS)
def validate_cap_mbgp_vpnv4fs(cmv4fs):
    if not isinstance(cmv4fs, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP '
                              'VPNv4 Flow Specification capability '
                              'settings: %s. Boolean value expected' % cmv4fs)
    return cmv4fs


@validate(name=CAP_MBGP_VPNV6FS)
def validate_cap_mbgp_vpnv66fs(cmv6fs):
    if not isinstance(cmv6fs, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP '
                              'VPNv6 Flow Specification capability '
                              'settings: %s. Boolean value expected' % cmv6fs)
    return cmv6fs


@validate(name=CAP_MBGP_L2VPNFS)
def validate_cap_mbgp_l2vpnfs(cml2fs):
    if not isinstance(cml2fs, bool):
        raise ConfigTypeError(desc='Invalid MP-BGP '
                              'L2VPN Flow Specification capability '
                              'settings: %s. Boolean value expected' % cml2fs)
    return cml2fs


@validate(name=CAP_RTC)
def validate_cap_rtc(cap_rtc):
    if not isinstance(cap_rtc, bool):
        raise ConfigTypeError(desc='Invalid type for specifying RTC '
                              'capability. Expected boolean got: %s' %
                              type(cap_rtc))
    return cap_rtc


@validate(name=RTC_AS)
def validate_cap_rtc_as(rtc_as):
    if not is_valid_asn(rtc_as):
        raise ConfigValueError(desc='Invalid RTC AS configuration value: %s'
                               % rtc_as)
    return rtc_as


@validate(name=HOLD_TIME)
def validate_hold_time(hold_time):
    if ((hold_time is None) or (not isinstance(hold_time, int)) or
            hold_time < 10):
        raise ConfigValueError(desc='Invalid hold_time configuration value %s'
                               % hold_time)

    return hold_time


@validate(name=MULTI_EXIT_DISC)
def validate_med(med):
    if med is not None and not validation.is_valid_med(med):
        raise ConfigValueError(desc='Invalid multi-exit-discriminatory (med)'
                               ' value: %s.' % med)
    return med


@validate(name=SITE_OF_ORIGINS)
def validate_soo_list(soo_list):
    if not isinstance(soo_list, list):
        raise ConfigTypeError(conf_name=SITE_OF_ORIGINS, conf_value=soo_list)
    if len(soo_list) > MAX_NUM_SOO:
        raise ConfigValueError(desc='Max. SOO is limited to %s' %
                               MAX_NUM_SOO)
    if not all(validation.is_valid_ext_comm_attr(attr) for attr in soo_list):
        raise ConfigValueError(conf_name=SITE_OF_ORIGINS,
                               conf_value=soo_list)
    # Check if we have duplicates
    unique_rts = set(soo_list)
    if len(unique_rts) != len(soo_list):
        raise ConfigValueError(desc='Duplicate value provided in %s' %
                               soo_list)
    return soo_list


@validate(name=MAX_PREFIXES)
def validate_max_prefixes(max_prefixes):
    if not isinstance(max_prefixes, int):
        raise ConfigTypeError(desc='Max. prefixes value should be of type '
                              'int or long but found %s' % type(max_prefixes))
    if max_prefixes < 0:
        raise ConfigValueError(desc='Invalid max. prefixes value: %s' %
                               max_prefixes)
    return max_prefixes


@validate(name=ADVERTISE_PEER_AS)
def validate_advertise_peer_as(advertise_peer_as):
    if not isinstance(advertise_peer_as, bool):
        raise ConfigTypeError(desc='Invalid type for advertise-peer-as, '
                              'expected bool got %s' %
                              type(advertise_peer_as))
    return advertise_peer_as


# =============================================================================
# Other utils.
# =============================================================================

def compute_optional_conf(conf_name, default_value, **all_config):
    """Returns *conf_name* settings if provided in *all_config*, else returns
     *default_value*.

    Validates *conf_name* value if provided.
    """
    conf_value = all_config.get(conf_name)
    if conf_value is not None:
        # Validate configuration value.
        conf_value = get_validator(conf_name)(conf_value)
    else:
        conf_value = default_value
    return conf_value



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2014 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import oslo_config.cfg

# there are 3 ways to access the configuration.
#
#    a. os_ken.cfg.CONF  (used to register cli options)
#    b. OSKenApp.CONF  (preferred way for os_ken applications)
#    c. oslo_config.cfg.CONF
#
# Currently a. and b. shares a single ConfigOpts instance.
# We intentionally avoid using c. for our options as a python program
# which embeds os_ken applications (eg. neutron agent) might want to put
# its own set of cli options into it, which can conflict with ours.
# (Currently there seems no conflict for the neutron agent.  But who knows?)
# At some point later we might want to unshare a. and b. as well, in order
# to allow app-specific options.

CONF = oslo_config.cfg.ConfigOpts()

# re-export for convenience

from oslo_config.cfg import ConfigOpts

from oslo_config.cfg import Opt
from oslo_config.cfg import BoolOpt
from oslo_config.cfg import IntOpt
from oslo_config.cfg import ListOpt
from oslo_config.cfg import MultiStrOpt
from oslo_config.cfg import StrOpt
from oslo_config.cfg import FloatOpt

from oslo_config.cfg import RequiredOptError
from oslo_config.cfg import ConfigFilesNotFoundError



import logging
import traceback

from os_ken.lib.packet.bgp import RouteFamily
from os_ken.lib.packet.bgp import RF_IPv4_UC
from os_ken.lib.packet.bgp import RF_IPv6_UC
from os_ken.lib.packet.bgp import RF_IPv4_VPN
from os_ken.lib.packet.bgp import RF_IPv6_VPN
from os_ken.lib.packet.bgp import RF_L2_EVPN
from os_ken.lib.packet.bgp import RF_IPv4_FLOWSPEC
from os_ken.lib.packet.bgp import RF_IPv6_FLOWSPEC
from os_ken.lib.packet.bgp import RF_VPNv4_FLOWSPEC
from os_ken.lib.packet.bgp import RF_VPNv6_FLOWSPEC
from os_ken.lib.packet.bgp import RF_L2VPN_FLOWSPEC
from os_ken.lib.packet.bgp import RF_RTC_UC
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_ORIGIN
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_AS_PATH
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_MULTI_EXIT_DISC
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_LOCAL_PREF
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_IGP
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_EGP
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_INCOMPLETE

from os_ken.services.protocols.bgp.base import add_bgp_error_metadata
from os_ken.services.protocols.bgp.base import BGPSException
from os_ken.services.protocols.bgp.base import SUPPORTED_GLOBAL_RF
from os_ken.services.protocols.bgp.core_manager import CORE_MANAGER


LOG = logging.getLogger('bgpspeaker.operator.internal_api')

INTERNAL_API_ERROR = 100
INTERNAL_API_SUB_ERROR = 101


class InternalApi(object):

    def __init__(self, log_handler=None):
        self.log_handler = log_handler

    def count_all_vrf_routes(self):
        vrf_tables = self._get_vrf_tables()
        ret = {}
        for key in vrf_tables.keys():
            vrf_name, vrf_rf = key
            ret.update(self.count_single_vrf_routes(vrf_name, vrf_rf))
        return ret

    def count_single_vrf_routes(self, vrf_name, vrf_rf):
        vrf = self._get_vrf_tables().get((vrf_name, vrf_rf))
        if vrf is None:
            raise WrongParamError('wrong vpn key %s' % str((vrf_name, vrf_rf)))
        vrf_name = vrf_name.encode('ascii', 'ignore')

        route_count = \
            len([d for d in vrf.values() if d.best_path])
        return {str((vrf_name, vrf_rf)): route_count}

    def get_vrfs_conf(self):
        return CORE_MANAGER.vrfs_conf.vrfs_by_rd_rf_id

    def get_all_vrf_routes(self):
        vrfs = self._get_vrf_tables()
        ret = {}
        for (vrf_id, vrf_rf), table in sorted(vrfs.items()):
            ret[str((vrf_id, vrf_rf))] = self._get_single_vrf_routes(table)
        return ret

    def get_single_vrf_routes(self, vrf_id, vrf_rf):
        vrf = self._get_vrf_table(vrf_id, vrf_rf)
        if not vrf:
            raise WrongParamError('wrong vpn name %s' % str((vrf_id, vrf_rf)))
        return [self._dst_to_dict(d) for d in vrf.values()]

    def _get_single_vrf_routes(self, vrf_table):
        return [self._dst_to_dict(d) for d in vrf_table.values()]

    def _get_vrf_table(self, vrf_name, vrf_rf):
        return CORE_MANAGER.get_core_service()\
            .table_manager.get_vrf_table(vrf_name, vrf_rf)

    def _get_vrf_tables(self):
        return CORE_MANAGER.get_core_service().table_manager.get_vrf_tables()

    def get_single_rib_routes(self, addr_family):
        rfs = {
            'ipv4': RF_IPv4_UC,
            'ipv6': RF_IPv6_UC,
            'vpnv4': RF_IPv4_VPN,
            'vpnv6': RF_IPv6_VPN,
            'evpn': RF_L2_EVPN,
            'ipv4fs': RF_IPv4_FLOWSPEC,
            'ipv6fs': RF_IPv6_FLOWSPEC,
            'vpnv4fs': RF_VPNv4_FLOWSPEC,
            'vpnv6fs': RF_VPNv6_FLOWSPEC,
            'l2vpnfs': RF_L2VPN_FLOWSPEC,
            'rtfilter': RF_RTC_UC
        }
        if addr_family not in rfs:
            raise WrongParamError('Unknown or unsupported family: %s' %
                                  addr_family)

        rf = rfs.get(addr_family)
        table_manager = self.get_core_service().table_manager
        gtable = table_manager.get_global_table_by_route_family(rf)
        if gtable is not None:
            return [self._dst_to_dict(dst)
                    for dst in sorted(gtable.values())]
        else:
            return []

    def _dst_to_dict(self, dst):
        ret = {'paths': [],
               'prefix': dst.nlri_str}

        def _path_to_dict(dst, path):

            path_seg_list = path.get_pattr(BGP_ATTR_TYPE_AS_PATH).path_seg_list

            if isinstance(path_seg_list, list):
                aspath = []
                for as_path_seg in path_seg_list:
                    for as_num in as_path_seg:
                        aspath.append(as_num)
            else:
                aspath = ''

            origin = path.get_pattr(BGP_ATTR_TYPE_ORIGIN)
            origin = origin.value if origin else None

            if origin == BGP_ATTR_ORIGIN_IGP:
                origin = 'i'
            elif origin == BGP_ATTR_ORIGIN_EGP:
                origin = 'e'
            elif origin == BGP_ATTR_ORIGIN_INCOMPLETE:
                origin = '?'

            nexthop = path.nexthop
            # Get the MED path attribute
            med = path.get_pattr(BGP_ATTR_TYPE_MULTI_EXIT_DISC)
            med = med.value if med else ''
            # Get best path reason
            bpr = dst.best_path_reason if path == dst.best_path else ''

            # Get local preference
            localpref = path.get_pattr(BGP_ATTR_TYPE_LOCAL_PREF)
            localpref = localpref.value if localpref else ''

            if hasattr(path.nlri, 'label_list'):
                labels = path.nlri.label_list
            else:
                labels = None

            return {'best': (path == dst.best_path),
                    'bpr': bpr,
                    'prefix': path.nlri_str,
                    'labels': labels,
                    'nexthop': nexthop,
                    'metric': med,
                    'aspath': aspath,
                    'origin': origin,
                    'localpref': localpref}

        for path in dst.known_path_list:
            ret['paths'].append(_path_to_dict(dst, path))

        return ret

    def check_logging(self):
        return self.log_handler and self._has_log_handler(self.log_handler)

    def check_logging_level(self):
        return logging.getLevelName(self.log_handler.level)

    def _has_log_handler(self, log_handler):
        if log_handler in logging.getLogger('bgpspeaker').handlers:
            return True
        return False

    def route_refresh(self, peer_ip=None, afi=None, safi=None):
        if not peer_ip:
            peer_ip = 'all'

        try:
            route_families = []
            if afi is None and safi is None:
                route_families.extend(SUPPORTED_GLOBAL_RF)
            else:
                route_family = RouteFamily(afi, safi)
                if route_family not in SUPPORTED_GLOBAL_RF:
                    raise WrongParamError('Not supported address-family'
                                          ' %s, %s' % (afi, safi))
                route_families.append(route_family)

            pm = CORE_MANAGER.get_core_service().peer_manager
            pm.make_route_refresh_request(peer_ip, *route_families)
        except Exception as e:
            LOG.error(traceback.format_exc())
            raise WrongParamError(str(e))
        return None

    def get_core_service(self):
        return CORE_MANAGER.get_core_service()


@add_bgp_error_metadata(code=INTERNAL_API_ERROR,
                        sub_code=INTERNAL_API_SUB_ERROR,
                        def_desc='Unknown internal api exception.')
class WrongParamError(BGPSException):
    pass
